from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Union

from clipped.compact.pydantic import (
    StrictFloat,
    StrictInt,
    StrictStr,
    strict_str_validator,
)
from clipped.config.constants import PARAM_REGEX

if TYPE_CHECKING:
    from clipped.compact.pydantic import CallableGenerator


class RefField(StrictStr):
    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield strict_str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value, **kwargs):
        if not isinstance(value, str):
            return value

        field = kwargs.get("field")
        param = PARAM_REGEX.search(value)
        if not param:  # TODO: Fix error message
            raise ValueError(
                f"Field `{field.name}` value must be equal to `{field.default}`, received `{value}` instead."
            )
        return value


BoolOrRef = Union[bool, RefField]
IntOrRef = Union[StrictInt, RefField]
StrictFloatOrRef = Union[StrictFloat, RefField]
FloatOrRef = Union[float, RefField]
DatetimeOrRef = Union[datetime, RefField]
TimeDeltaOrRef = Union[timedelta, RefField]



import functools
import os
import pprint

from collections.abc import Mapping
from typing import Any, Callable, Dict, List, Optional, Set, Type, Union

from clipped.compact.pydantic import BaseModel, Extra, create_model
from clipped.config.exceptions import SchemaError
from clipped.config.patch_strategy import PatchStrategy
from clipped.config.spec import ConfigSpec
from clipped.utils.dicts import deep_update
from clipped.utils.humanize import humanize_timesince
from clipped.utils.json import orjson_dumps, orjson_loads
from clipped.utils.strings import to_camel_case
from clipped.utils.units import to_percentage, to_unit_memory


class BaseSchemaModel(BaseModel):
    _IDENTIFIER: Optional[str] = None
    _DEFAULT_INCLUDE_ATTRIBUTES = []
    _DEFAULT_EXCLUDE_ATTRIBUTES = []
    _DATETIME_ATTRIBUTES = []
    _MEM_SIZE_ATTRIBUTES = []
    _PERCENT_ATTRIBUTES = []
    _ROUNDING = 2
    _WRITE_MODE = 0o777
    _FIELDS_MANUAL_PATCH = []
    _FIELDS_SAME_KIND_PATCH = []
    _SWAGGER_FIELDS = []
    _SWAGGER_FIELDS_LISTS = ["tolerations"]
    _PARTIAL = False
    _VERSION = None
    _CONFIG_SPEC = ConfigSpec
    _SCHEMA_EXCEPTION = SchemaError
    _USE_DISCRIMINATOR = False

    def __init__(self, **data):
        if self._USE_DISCRIMINATOR:
            data["kind"] = data.pop("kind", self._IDENTIFIER)
        super().__init__(**data)

    @classmethod
    def get_identifier(cls):
        return cls._IDENTIFIER

    def to_light_dict(
        self,
        humanize_values: bool = False,
        include_attrs: List[str] = None,
        exclude_attrs: List[str] = None,
    ) -> Dict[str, Any]:
        obj_dict = self.to_dict(humanize_values=humanize_values)
        if all([include_attrs, exclude_attrs]):
            raise self._SCHEMA_EXCEPTION(
                "Only one value `include_attrs` or `exclude_attrs` is allowed."
            )
        if not any([include_attrs, exclude_attrs]):  # Use Default setup attrs
            include_attrs = self._DEFAULT_INCLUDE_ATTRIBUTES
            exclude_attrs = self._DEFAULT_EXCLUDE_ATTRIBUTES

        if include_attrs:
            exclude_attrs = set(obj_dict.keys()) - set(include_attrs)
        for attr in exclude_attrs:
            obj_dict.pop(attr, None)
        return obj_dict

    def to_dict(
        self,
        humanize_values: bool = False,
        include_kind: bool = False,
        include_version: bool = False,
        exclude_unset: bool = True,
        exclude_none: bool = True,
        exclude_defaults: bool = False,
    ) -> Dict[str, Any]:
        obj = self.obj_to_dict(
            self,
            humanize_values=humanize_values,
            include_kind=include_kind,
            include_version=include_version,
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )
        return obj

    def to_json(
        self,
        humanize_values: bool = False,
        include_kind: bool = False,
        include_version: bool = False,
    ) -> str:
        if include_kind and "kind" in self.__fields__.keys():
            self.kind = self._IDENTIFIER

        if include_version and "version" in self.__fields__.keys():
            self.version = self._VERSION
        return self.obj_to_json(
            self,
            humanize_values=humanize_values,
        )

    def to_str(self) -> str:
        return pprint.pformat(self.dict(by_alias=True))

    def to_schema(self) -> Dict:
        return self.obj_to_schema(self)

    @classmethod
    def humanize_attrs(cls, obj) -> Dict:
        humanized_attrs = {}
        for attr in cls._DATETIME_ATTRIBUTES:
            humanized_attrs[attr] = humanize_timesince(getattr(obj, attr))
        for attr in cls._PERCENT_ATTRIBUTES:
            humanized_attrs[attr] = to_percentage(getattr(obj, attr), cls._ROUNDING)
        for attr in cls._MEM_SIZE_ATTRIBUTES:
            humanized_attrs[attr] = to_unit_memory(getattr(obj, attr))
        return humanized_attrs

    @classmethod
    def obj_to_json(
        cls,
        obj,
        humanize_values: bool = False,
        exclude_unset: bool = True,
        exclude_none: bool = True,
        exclude_defaults: bool = False,
    ) -> str:
        return obj.json(
            by_alias=True,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    @staticmethod
    def swagger_to_dict(fdata: Any) -> Union[Dict, List[Dict]]:
        def _field_to_dict(fvalue: Any):
            if hasattr(fvalue, "to_dict"):
                fvalue = fvalue.to_dict()
                keys = set(fvalue.keys())
                for k in keys:
                    if fvalue[k] is None:
                        del fvalue[k]
                return {to_camel_case(k): fvalue[k] for k in fvalue}
            return fvalue

        if not fdata:
            return fdata
        elif fdata and isinstance(fdata, list):
            return [_field_to_dict(v) for v in fdata]
        else:
            return _field_to_dict(fdata)

    def dict(self, *args, **kwargs) -> Dict:
        data = super().dict(*args, **kwargs)
        # Handle swagger fields and perform `to_dict`
        for field in self._SWAGGER_FIELDS:
            if field in data:
                data[field] = self.swagger_to_dict(data[field])
        return data

    @classmethod
    def obj_to_dict(
        cls,
        obj: "BaseSchemaModel",
        humanize_values: bool = False,
        include_kind: bool = False,
        include_version: bool = False,
        exclude_unset: bool = True,
        exclude_none: bool = True,
        exclude_defaults: bool = False,
    ) -> Dict:
        humanized_attrs = cls.humanize_attrs(obj) if humanize_values else {}
        data_dict = obj.dict(
            by_alias=True,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        if include_kind and "kind" not in data_dict and hasattr(obj, "kind"):
            data_dict["kind"] = obj._IDENTIFIER

        if include_version and "version" not in data_dict:
            data_dict["version"] = cls._VERSION

        for k, v in humanized_attrs.items():
            data_dict[k] = v
        return data_dict

    @classmethod
    def obj_to_schema(cls, obj: "BaseSchemaModel") -> Dict:
        return {cls._IDENTIFIER: cls.obj_to_dict(obj)}

    @classmethod
    def from_dict(cls, value: Any, partial: bool = False) -> "BaseSchemaModel":
        return cls.parse_obj(value)

    @classmethod
    def read(
        cls, values: Any, partial: bool = False, config_type: str = None
    ) -> "BaseSchemaModel":
        values = cls._CONFIG_SPEC.read_from(values, config_type=config_type)
        return cls.from_dict(values, partial=partial)

    @classmethod
    def init_file(cls, filepath: str, config: Optional["BaseSchemaModel"] = None):
        if not os.path.exists(filepath):
            cls.write(config or cls(), filepath=filepath, mode=cls._WRITE_MODE)

    def write(self, filepath: str, mode: Optional[int] = None):
        with open(filepath, "w") as config_file:
            config_file.write(self.to_json())
            if mode is not None:
                os.chmod(filepath, mode)

    def clone(self):
        return self.from_dict(self.to_dict())

    @staticmethod
    def patch_normal_merge(
        current_value: Any, value: Any, strategy: Optional[PatchStrategy] = None
    ) -> Any:
        strategy = strategy or PatchStrategy.POST_MERGE

        if isinstance(current_value, dict):
            if PatchStrategy.is_post_merge(strategy):
                return deep_update(current_value, value)
            elif PatchStrategy.is_pre_merge(strategy):
                return deep_update(value, current_value)
        elif isinstance(current_value, list):
            if PatchStrategy.is_post_merge(strategy):
                return current_value + [i for i in value if i not in current_value]
            elif PatchStrategy.is_pre_merge(strategy):
                return value + [i for i in current_value if i not in value]
        elif isinstance(current_value, BaseSchemaModel):
            return current_value.patch(value, strategy=strategy)
        elif hasattr(current_value, "to_dict"):
            if PatchStrategy.is_post_merge(strategy):
                return deep_update(current_value.to_dict(), value.to_dict())
            elif PatchStrategy.is_pre_merge(strategy):
                return deep_update(value.to_dict(), current_value.to_dict())
        else:
            if PatchStrategy.is_post_merge(strategy):
                return value
            elif PatchStrategy.is_pre_merge(strategy):
                return current_value

    @classmethod
    def patch_swagger_field(
        cls,
        config: "BaseSchemaModel",
        values: Dict,
        strategy: Optional[PatchStrategy] = None,
    ):
        strategy = strategy or PatchStrategy.POST_MERGE

        openapi_types = getattr(config, "openapi_types", {})
        for key in openapi_types:
            value = getattr(values, key, None)
            if value is None:
                continue

            current_value = getattr(config, key, None)
            if current_value is None:
                setattr(
                    config, key, value
                )  # handles also PatchStrategy.ISNULL implicitly
                continue

            if PatchStrategy.is_null(strategy):
                continue
            if PatchStrategy.is_replace(strategy):
                setattr(config, key, value)
                continue

            setattr(config, key, cls.patch_normal_merge(current_value, value, strategy))

    @classmethod
    def patch_swagger_field_list(
        cls, current_value: Any, value: any, strategy: Optional[PatchStrategy] = None
    ) -> Any:
        strategy = strategy or PatchStrategy.POST_MERGE

        if PatchStrategy.is_null(strategy):
            return current_value
        if PatchStrategy.is_replace(strategy):
            return value

        return cls.patch_normal_merge(current_value, value, strategy)

    @classmethod
    def patch_obj(
        cls,
        config: Union[Dict, "BaseSchemaModel"],
        values: Union[Dict, "BaseSchemaModel"],
        strategy: Optional[PatchStrategy] = None,
    ):
        strategy = strategy or PatchStrategy.POST_MERGE

        for key in config.__fields__.keys():
            if key in cls._FIELDS_MANUAL_PATCH:
                continue

            value = getattr(values, key, None)
            if value is None:
                should_continue = True
                if isinstance(values, Mapping) and key in values:
                    should_continue = False
                elif (
                    isinstance(values, BaseSchemaModel) and key in values.__fields_set__
                ):
                    should_continue = False

                if should_continue:
                    continue

            current_value = getattr(config, key, None)
            if current_value is None:
                setattr(
                    config, key, value
                )  # handles also PatchStrategy.ISNULL implicitly
                continue

            if (
                isinstance(current_value, BaseSchemaModel)
                and key not in cls._FIELDS_SAME_KIND_PATCH
            ):
                current_value.patch(value, strategy=strategy)
                continue

            if not isinstance(current_value, BaseSchemaModel) and (
                hasattr(current_value, "openapi_types") or key in cls._SWAGGER_FIELDS
            ):
                if key in cls._SWAGGER_FIELDS_LISTS:  # Special case for lists
                    setattr(
                        config,
                        key,
                        cls.patch_swagger_field_list(current_value, value, strategy),
                    )
                else:
                    cls.patch_swagger_field(current_value, value, strategy)
                continue

            if PatchStrategy.is_null(strategy):
                continue
            if PatchStrategy.is_replace(strategy):
                setattr(config, key, value)
                continue

            # We only handle merge strategies
            def normal_merge():
                setattr(
                    config, key, cls.patch_normal_merge(current_value, value, strategy)
                )

            def same_kind_merge():
                # If the same kind resume merge patch using base logic
                if value and current_value.kind == value.kind:
                    normal_merge()
                # Not same kind use post/pre replace
                else:
                    if PatchStrategy.is_post_merge(strategy):
                        setattr(config, key, value)
                    elif PatchStrategy.is_pre_merge(strategy):
                        setattr(config, key, current_value)

            if key in cls._FIELDS_SAME_KIND_PATCH or value is None:
                same_kind_merge()
            else:
                normal_merge()

        return config

    def patch(
        self,
        values: Union[Dict, "BaseSchemaModel"],
        strategy: Optional[PatchStrategy] = None,
    ):
        strategy = strategy or PatchStrategy.POST_MERGE
        return self.patch_obj(self, values, strategy)

    @classmethod
    def get_keys_and_aliases(cls) -> Dict[str, str]:
        return {key: field.alias for key, field in cls.__fields__.items()}

    @classmethod
    def get_all_possible_keys(cls) -> Set[str]:
        keys_and_aliases = cls.get_keys_and_aliases()
        all_keys = {key for key, value in keys_and_aliases.items()}
        all_aliases = {value for key, value in keys_and_aliases.items()}
        return all_keys | all_aliases

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True
        arbitrary_types_allowed = True
        use_enum_values = True
        extra = Extra.forbid
        json_dumps: Callable[..., str] = orjson_dumps
        json_loads = orjson_loads


class BaseAllowSchemaModel(BaseSchemaModel):
    class Config(BaseSchemaModel.Config):
        extra = Extra.allow


def base_schema(cls, new_cls: Type[BaseSchemaModel]):
    """
    `base_schema` is a decorator to add the `BaseSchemaModel` to a class.

    usage example:
        @base_schema
        class MySchema:
            ...
    """

    if not issubclass(cls, BaseModel):
        raise TypeError("Class must be a subclass of pydantic.BaseModel")

    return functools.wraps(cls, updated=())(
        create_model(
            cls.__name__,
            __base__=(cls, new_cls),
        )
    )  # type: ignore


def to_partial(cls):
    class NewCls(cls):
        _PARTIAL = True

    NewCls.__name__ = f"Partial{cls.__name__}"

    for field in NewCls.__fields__.values():
        if hasattr(field.type_, "__base__") and issubclass(
            field.type_.__base__, BaseModel
        ):
            field.type_ = to_partial(field.type_)

    return NewCls


def skip_partial(f):
    """
    `skip_partial` is a decorator to skip validation when `_PARTIAL = True` and return the data as is.

    usage example:
        @root_validator
        @skip_partial
        def my_custom_check(cls, values):
            ...
            return ...
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        cls_arg = None
        v_args = None
        if args:
            cls_arg = args[0]
            if len(args) > 1:
                v_args = args[1]
        partial_property = getattr(cls_arg, f"_PARTIAL", None)
        if partial_property:
            return v_args
        return f(*args, **kwargs)

    return wrapper



import logging
import os

from collections.abc import Mapping
from typing import Any, Dict, Optional, Type

from clipped.config.contexts import get_project_path, get_temp_path
from clipped.config.reader import ConfigReader
from clipped.config.schema import BaseSchemaModel
from clipped.utils.enums import PEnum
from clipped.utils.json import orjson_dumps, orjson_loads
from clipped.utils.paths import check_or_create_path

_logger = logging.getLogger("clipped.config.manager")


class ConfigManager:
    """Base class for managing a configuration file."""

    class Visibility(str, PEnum):
        GLOBAL = "global"
        LOCAL = "local"
        ALL = "all"

    VISIBILITY: Visibility = None
    IN_PROJECT_DIR = False
    CONFIG_PATH: Optional[str] = None
    CONFIG_FILE_NAME: Optional[str] = None
    CONFIG: Type[BaseSchemaModel] = None
    _CONFIG_READER: Type[ConfigReader] = ConfigReader
    _LOGGER = _logger
    _PROJECT = ".clipped"
    _PROJECT_PATH: str = None
    _TEMP_PATH: str = None

    @classmethod
    def _get_project_path(cls) -> str:
        if cls._PROJECT_PATH:
            return cls._PROJECT_PATH

        cls._PROJECT_PATH = get_project_path(cls._PROJECT)
        return cls._PROJECT_PATH

    @classmethod
    def _get_temp_path(cls) -> str:
        if cls._TEMP_PATH:
            return cls._TEMP_PATH

        cls._TEMP_PATH = get_temp_path(cls._PROJECT)
        return cls._TEMP_PATH

    @classmethod
    def is_global(cls, visibility: Optional[Visibility] = None) -> bool:
        visibility = visibility or cls.VISIBILITY
        return visibility == cls.Visibility.GLOBAL

    @classmethod
    def is_local(cls, visibility=None) -> bool:
        visibility = visibility or cls.VISIBILITY
        return visibility == cls.Visibility.LOCAL

    @classmethod
    def is_all_visibility(cls, visibility=None) -> bool:
        visibility = visibility or cls.VISIBILITY
        return visibility == cls.Visibility.ALL

    @classmethod
    def get_visibility(cls) -> str:
        if cls.is_all_visibility():
            return (
                cls.Visibility.LOCAL
                if cls.is_locally_initialized()
                else cls.Visibility.GLOBAL
            )
        return cls.VISIBILITY

    @classmethod
    def set_config_path(cls, config_path: Optional[str]):
        cls.CONFIG_PATH = config_path

    @classmethod
    def _create_dir(cls, config_file_path):
        try:
            check_or_create_path(config_file_path, is_dir=False)
        except OSError:
            # Except permission denied and potential race conditions
            # in multi-threaded environments.
            cls._LOGGER.error(
                "Could not create config context directory for file `%s`",
                config_file_path,
            )

    @staticmethod
    def _get_and_check_path(config_path: str) -> Optional[str]:
        if config_path and os.path.isfile(config_path):
            return config_path
        return None

    @classmethod
    def get_local_config_path(cls) -> str:
        # local to this directory
        base_path = os.path.join(".")
        if cls.IN_PROJECT_DIR:
            # Add it to the current ".project" path
            base_path = os.path.join(base_path, cls._PROJECT)
        config_path = os.path.join(base_path, cls.CONFIG_FILE_NAME)
        return config_path

    @classmethod
    def check_local_config_path(cls) -> Optional[str]:
        return cls._get_and_check_path(cls.get_local_config_path())

    @classmethod
    def get_global_config_path(cls) -> str:
        if cls.CONFIG_PATH:
            base_path = os.path.join(cls.CONFIG_PATH, cls._PROJECT)
        else:
            base_path = cls._get_project_path()
        config_path = os.path.join(base_path, cls.CONFIG_FILE_NAME)
        return config_path

    @classmethod
    def check_global_config_path(cls) -> Optional[str]:
        return cls._get_and_check_path(cls.get_global_config_path())

    @classmethod
    def get_tmp_config_path(cls) -> str:
        base_path = cls._get_temp_path()
        config_path = os.path.join(base_path, cls.CONFIG_FILE_NAME)
        return config_path

    @classmethod
    def get_config_filepath(
        cls, create: bool = True, visibility: Optional[str] = None
    ) -> str:
        config_path = None
        if cls.is_local(visibility):
            config_path = cls.get_local_config_path()
        elif cls.is_global(visibility):
            config_path = cls.get_global_config_path()
        elif cls.is_all_visibility(visibility):
            config_path = cls.check_local_config_path()
            if not config_path:
                config_path = cls.get_global_config_path()

        if create and config_path:
            cls._create_dir(config_path)
        return config_path

    @classmethod
    def init_config(cls, visibility: Optional[str] = None):
        config = cls.get_config()
        cls.set_config(config, init=True, visibility=visibility)

    @classmethod
    def is_locally_initialized(cls) -> Optional[str]:
        return cls.check_local_config_path()

    @classmethod
    def is_initialized(cls) -> Optional[str]:
        return cls._get_and_check_path(cls.get_config_filepath(create=False))

    @classmethod
    def set_config(
        cls, config: Any, init: bool = False, visibility: Optional[str] = None
    ):
        config_filepath = cls.get_config_filepath(visibility=visibility)

        if os.path.isfile(config_filepath) and init:
            cls._LOGGER.debug(
                "%s file already present at %s\n", cls.CONFIG_FILE_NAME, config_filepath
            )
            return

        with open(config_filepath, "w") as config_file:
            if hasattr(config, "to_json"):
                cls._LOGGER.debug(
                    "Setting %s in the file %s\n",
                    config.to_dict(),
                    cls.CONFIG_FILE_NAME,
                )
                config_file.write(config.to_json())
            elif hasattr(config, "to_dict"):
                cls._LOGGER.debug(
                    "Setting %s in the file %s\n",
                    config.to_dict(),
                    cls.CONFIG_FILE_NAME,
                )
                config_file.write(config.to_dict())
            elif isinstance(config, Mapping):
                config_file.write(orjson_dumps(config))
            else:
                cls._LOGGER.debug(
                    "Setting %s in the file %s\n", config, cls.CONFIG_FILE_NAME
                )
                config_file.write(config)

    @classmethod
    def get_config(cls, check: bool = True) -> Optional[Any]:
        if check and not cls.is_initialized():
            return None

        config_filepath = cls.get_config_filepath()
        cls._LOGGER.debug(
            "Reading config `%s` from path: %s\n", cls.__name__, config_filepath
        )
        return cls.read_from_path(config_filepath)

    @classmethod
    def read_from_path(cls, config_filepath: str) -> Optional[Any]:
        if issubclass(cls.CONFIG, BaseSchemaModel):
            return cls.CONFIG.read(config_filepath)
        with open(config_filepath, "r") as config_file:
            config_str = config_file.read()
        return cls.CONFIG(**orjson_loads(config_str))

    @classmethod
    def get_config_defaults(cls) -> Dict:
        return {}

    @classmethod
    def get_config_or_default(cls) -> Any:
        if not cls.is_initialized():
            return cls.CONFIG(
                **cls.get_config_defaults()
            )  # pylint:disable=not-callable

        return cls.get_config(check=False)

    @classmethod
    def get_config_from_env(cls, **kwargs) -> Any:
        raise NotImplementedError

    @classmethod
    def get_value(cls, key) -> Optional[Any]:
        config = cls.get_config()
        if config:
            if hasattr(config, key):
                return getattr(config, key)
            else:
                cls._LOGGER.warning(
                    "Config `%s` has no key `%s`", cls.CONFIG.__name__, key
                )

        return None

    @classmethod
    def purge(cls, visibility: Optional[str] = None):
        def _purge():
            if config_filepath and os.path.isfile(config_filepath):
                os.remove(config_filepath)

        if cls.is_all_visibility():
            if visibility:
                config_filepath = cls.get_config_filepath(
                    create=False, visibility=visibility
                )
                _purge()
            else:
                config_filepath = cls.get_config_filepath(
                    create=False, visibility=cls.Visibility.LOCAL
                )
                _purge()
                config_filepath = cls.get_config_filepath(
                    create=False, visibility=cls.Visibility.GLOBAL
                )
                _purge()
        else:
            config_filepath = cls.get_config_filepath(create=False)
            _purge()



from click import ClickException


class ClippedException(ClickException):
    def __init__(self, message=None):
        super().__init__(message)

    def __repr__(self):
        return self.message


class SchemaError(ClippedException):
    pass



from typing import TYPE_CHECKING

from pydantic.version import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import (
        AnyUrl,
        BaseConfig,
        BaseModel,
        ByteSize,
        ConstrainedBytes,
        ConstrainedDate,
        ConstrainedDecimal,
        ConstrainedFloat,
        ConstrainedFrozenSet,
        ConstrainedInt,
        ConstrainedList,
        ConstrainedSet,
        ConstrainedStr,
        Extra,
        Field,
        FiniteFloat,
        FutureDate,
        Json,
        NegativeFloat,
        NegativeInt,
        NoneBytes,
        NoneStr,
        NoneStrBytes,
        NonNegativeFloat,
        NonNegativeInt,
        NonPositiveFloat,
        NonPositiveInt,
        PastDate,
        PaymentCardNumber,
        PositiveFloat,
        PositiveInt,
        PrivateAttr,
        PydanticTypeError,
        PydanticValueError,
        PyObject,
        SecretBytes,
        SecretField,
        SecretStr,
        StrBytes,
        StrictBool,
        StrictBytes,
        StrictFloat,
        StrictInt,
        StrictStr,
        ValidationError,
        constr,
        create_model,
        root_validator,
        validate_arguments,
        validator,
    )
    from pydantic.v1.datetime_parse import parse_date, parse_datetime, parse_duration
    from pydantic.v1.fields import ModelField
    from pydantic.v1.parse import Protocol, load_str_bytes
    from pydantic.v1.tools import NameFactory, _get_parsing_type
    from pydantic.v1.validators import strict_str_validator, uuid_validator

    if TYPE_CHECKING:
        from pydantic.v1.typing import CallableGenerator
else:
    from pydantic import (
        AnyUrl,
        BaseConfig,
        BaseModel,
        ByteSize,
        ConstrainedBytes,
        ConstrainedDate,
        ConstrainedDecimal,
        ConstrainedFloat,
        ConstrainedFrozenSet,
        ConstrainedInt,
        ConstrainedList,
        ConstrainedSet,
        ConstrainedStr,
        Extra,
        Field,
        FiniteFloat,
        FutureDate,
        Json,
        NegativeFloat,
        NegativeInt,
        NoneBytes,
        NoneStr,
        NoneStrBytes,
        NonNegativeFloat,
        NonNegativeInt,
        NonPositiveFloat,
        NonPositiveInt,
        PastDate,
        PaymentCardNumber,
        PositiveFloat,
        PositiveInt,
        PrivateAttr,
        PydanticTypeError,
        PydanticValueError,
        PyObject,
        SecretBytes,
        SecretField,
        SecretStr,
        StrBytes,
        StrictBool,
        StrictBytes,
        StrictFloat,
        StrictInt,
        StrictStr,
        ValidationError,
        constr,
        create_model,
        root_validator,
        validate_arguments,
        validator,
    )
    from pydantic.datetime_parse import parse_date, parse_datetime, parse_duration
    from pydantic.fields import ModelField
    from pydantic.parse import Protocol, load_str_bytes
    from pydantic.tools import NameFactory, _get_parsing_type
    from pydantic.validators import strict_str_validator, uuid_validator

    if TYPE_CHECKING:
        from pydantic.typing import CallableGenerator



import logging

from typing import Any, Callable, List, Optional, Type, TypeVar

from clipped.compact.pydantic import (
    NameFactory,
    Protocol,
    PydanticTypeError,
    PydanticValueError,
    StrBytes,
    ValidationError,
    _get_parsing_type,
    load_str_bytes,
)
from clipped.config.constants import NO_VALUE_FOUND
from clipped.config.exceptions import SchemaError
from clipped.decorators.memoization import memoize
from clipped.utils.json import orjson_loads

_logger = logging.getLogger("clipped.parser")

T = TypeVar("T")


class ConfigParser:
    _SCHEMA_EXCEPTION = SchemaError

    @staticmethod
    @memoize
    def is_loadable(type_: Any):
        from clipped import types

        return type_ not in types.NON_LOADABLE

    @staticmethod
    @memoize
    def type_mapping():
        from clipped import types

        return types.MAPPING

    @staticmethod
    @memoize
    def type_forwarding():
        from clipped import types

        return types.FORWARDING

    @classmethod
    def parse_obj_as(
        cls, type_: Type[T], obj: Any, *, type_name: Optional[NameFactory] = None
    ) -> T:
        model_type = _get_parsing_type(type_, type_name=type_name)
        model_type.update_forward_refs(**cls.type_forwarding())
        return model_type(__root__=obj).__root__

    @classmethod
    def parse_raw_as(
        cls,
        type_: Type[T],
        b: StrBytes,
        *,
        content_type: str = None,
        encoding: str = "utf8",
        proto: Protocol = None,
        allow_pickle: bool = False,
        json_loads: Callable[[str], Any] = orjson_loads,
        type_name: Optional[NameFactory] = None,
    ) -> T:
        obj = load_str_bytes(
            b,
            proto=proto,
            content_type=content_type,
            encoding=encoding,
            allow_pickle=allow_pickle,
            json_loads=json_loads,
        )
        return cls.parse_obj_as(type_, obj, type_name=type_name)

    @classmethod
    def _check_options(cls, key, value, options):
        if options and value not in options:
            raise cls._SCHEMA_EXCEPTION(
                "The value `{}` provided for key `{}` "
                "is not one of the possible values.".format(value, key)
            )

    @classmethod
    def _get_typed_value(
        cls,
        key,
        value,
        target_type,
        is_optional=False,
        default=None,
        options=None,
    ):
        """
        Return the value corresponding to the key converted to the given type.

        Args:
            key: the dict key.
            target_type: The type we expect the variable or key to be in.
            is_optional: To raise an error if key was not found.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
            The corresponding value of the key converted.
        """
        if value is None or value == NO_VALUE_FOUND:
            if not is_optional:
                raise cls._SCHEMA_EXCEPTION(
                    "No value was provided for the non optional key `{}`.".format(key)
                )
            return default

        try:
            if cls.is_loadable(target_type) and isinstance(value, str):
                new_value = cls.parse_raw_as(
                    target_type, value, json_loads=orjson_loads
                )
            else:
                new_value = cls.parse_obj_as(target_type, value)
            cls._check_options(key=key, value=new_value, options=options)
            return new_value
        except (
            TypeError,
            ValueError,
            ValidationError,
            PydanticTypeError,
            PydanticValueError,
        ):
            raise cls._SCHEMA_EXCEPTION(
                "Cannot convert value `{}` (key: `{}`) "
                "to `{}`".format(value, key, target_type)
            )

    @classmethod
    def parse(cls, target_type: Any) -> Callable:
        """
        Get the value corresponding to the key and converts it to `target_type` using `type_convert`.

        Args
            target_type: the type to parse the values to.
            type_convert: the converter to use to cast the values.
            base_types: Optional list of types to convert from if the type does not correspond to target type.

        Returns:
             `Callback`: A parser for specific for the target_type.
        """

        def _parse(
            key: Any,
            value: Any,
            is_list: bool = False,
            is_optional: bool = False,
            default: Any = None,
            options: List[Type] = None,
        ):
            """
            Get the value corresponding to the key and converts it to `target_type` using `type_convert`.

            Args
                key: the dict key.
                value: the value to parse.
                is_list: If this is one element or a list of elements.
                is_optional: To raise an error if key was not found.
                default: default value if is_optional is True.
                options: list/tuple if provided, the value must be one of these values.

            Returns:
                 `target_type`: value corresponding to the key.
            """
            _target_type = cls.type_mapping().get(target_type, target_type)

            def _get_name():
                if hasattr(_target_type, "_name"):
                    return _target_type._name
                if hasattr(_target_type, "__name__"):
                    return _target_type.__name__
                return "Any"

            if is_list:
                _target_type = (
                    _target_type if isinstance(_target_type, str) else _get_name()
                )
                _target_type = f"List[{_target_type}]"

            return cls._get_typed_value(
                key=key,
                value=value,
                target_type=_target_type,
                is_optional=is_optional,
                default=default,
                options=options,
            )

        return _parse



from typing import TYPE_CHECKING
from uuid import UUID

from clipped.utils.json import orjson_dumps

if TYPE_CHECKING:
    from clipped.compact.pydantic import CallableGenerator


class GenericStr(str):
    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(cls, value, **kwargs):
        base_types = (int, float, dict, list, tuple, set)
        if isinstance(value, base_types):
            return orjson_dumps(value)
        if isinstance(value, str):
            return value
        if isinstance(value, UUID):
            return value.hex
        if value is None:
            return value
        try:
            return str(value)
        except Exception as e:
            field = kwargs.get("field")
            raise TypeError(
                f"Field `{field.name}` value must be a valid str or a value that can be casted to a str, received `{value}` instead."
            ) from e



import re

from typing import TYPE_CHECKING

from clipped.compact.pydantic import StrictStr, strict_str_validator

if TYPE_CHECKING:
    from clipped.compact.pydantic import CallableGenerator


class EmailStr(StrictStr):
    USER_REGEX = re.compile(
        r"(^[-!#$%&'*+/=?^`{}|~\w]+(\.[-!#$%&'*+/=?^`{}|~\w]+)*\Z"  # dot-atom
        # quoted-string
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]'
        r'|\\[\001-\011\013\014\016-\177])*"\Z)',
        re.IGNORECASE | re.UNICODE,
    )

    DOMAIN_REGEX = re.compile(
        # domain
        r"(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+" r"(?:[A-Z]{2,6}|[A-Z0-9-]{2,})\Z"
        # literal form, ipv4 address (SMTP 4.1.3)
        r"|^\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)"
        r"(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]\Z",
        re.IGNORECASE | re.UNICODE,
    )

    DOMAIN_WHITELIST = ("localhost",)

    ERROR_MESSAGE = "Not a valid email address."

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield strict_str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value, **kwargs):
        if not isinstance(value, str):
            return value

        if not value or "@" not in value:
            raise ValueError(cls.ERROR_MESSAGE)

        user_part, domain_part = value.rsplit("@", 1)

        if not cls.USER_REGEX.match(user_part):
            raise ValueError(cls.ERROR_MESSAGE)

        if domain_part not in cls.DOMAIN_WHITELIST:
            if not cls.DOMAIN_REGEX.match(domain_part):
                try:
                    domain_part = domain_part.encode("idna").decode("ascii")
                except UnicodeError:
                    pass
                else:
                    if cls.DOMAIN_REGEX.match(domain_part):
                        return value
                raise ValueError(cls.ERROR_MESSAGE)

        return value
